// XADecompress.cpp: implementation of the XADecompress class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "XADecompress.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

float facttable1[4];
float facttable2[4];
long table[16][16];
bool __tables_inited=false;

void _inittables(void)
{
	if (__tables_inited) return;

	int offsets[16] = {0,1,2,3,4,5,6,7,-8,-7,-6,-5,-4,-3,-2,-1};

	facttable1[0] = 0.0f;
	facttable1[1] = 0.9375f;
	facttable1[2] = 1.79688f;
	facttable1[3] = 1.53125f;

	facttable2[0] = 0.0f;
	facttable2[1] = 0.0f;
	facttable2[2] = -0.8125f;
	facttable2[3] = -0.859375f;

	int fact = 4096;
	for (int i=0; i<16; i++)
	{
		for (int j=0; j<16; j++)
		{
			table[i][j] = offsets[j]*fact;
		}
		fact>>=1;
	}

	__tables_inited = true;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XADecompress::XADecompress()
{
	_inittables();
	Track = NULL;
}

XADecompress::~XADecompress()
{
	if (Track) delete[] Track;
}

void XADecompress::Init(BYTE *Input, DWORD InSize, CFile *Output)
{
	pMap = Input;
	Size = InSize;
	Pos = 24;
	output = Output;

	Tracks = pMap[10];
	if (Track) delete[] Track;
	Track = new XATrack[Tracks];
}

BOOL XADecompress::Decompress()
{
	if (Pos>Size) return FALSE;

	DWORD dwfmtSize=16, dwDataSize=((Size-24)-(Size-24)/15)*4;
	DWORD dwRIFFSize=dwfmtSize+8+dwDataSize+8+4;

	output->SetLength(0);

	output->Write("RIFF", 4);
	output->Write(&dwRIFFSize, 4);
	output->Write("WAVEfmt ", 8);
	output->Write(&dwfmtSize, 4);
	output->Write(&pMap[8], 16);
	output->Write("data", 4);
	output->Write(&dwDataSize, 4);

#define rangecheck(s) if (s>4) return FALSE;
#define gettoken(s) if (!buf.get(s)) return TRUE;
#define output(s, t) gettoken(s) t = Track[i].process(s); output->Write(&t, 2);

	XABuffer buf;
	buf.Init(pMap+0x18, Size-0x18);

	unsigned char token;
	unsigned short out;

	while (1)
	{
		for (DWORD i=0; i<Tracks; i++)
		{
			if (Track[i].pos%15==0)
			{
				gettoken(token);
				rangecheck(token);
				Track[i].factLast = facttable1[token];
				Track[i].factPreLast = facttable2[token];

				gettoken(token);
				Track[i].row = token;
			}
			else
			{
				output(token, out);
				output(token, out);
			}

			Track[i].pos++;
		}
	}
}

XADecompress::XATrack::XATrack()
{
	factLast = factPreLast = 0;
	Last = PreLast = 0;
	pos = 0;
	row = 0;
}